/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997-2012 Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    slouken@libsdl.org
*/
#include "SDL_config.h"

/*
	MiNT audio driver
	using XBIOS functions (Falcon)

	Patrice Mandin, Didier Méquignon
*/

#include <unistd.h>
#include <support.h>

/* Mint includes */
#include <mint/osbind.h>
#include <mint/falcon.h>
#include <mint/cookie.h>
#include <mint/sysvars.h>

#include "SDL_audio.h"
#include "../SDL_audio_c.h"
#include "../SDL_sysaudio.h"

#include "../../video/ataricommon/SDL_atarimxalloc_c.h"
#include "../../video/ataricommon/SDL_atarisuper.h"

#include "SDL_mintaudio.h"
#include "SDL_mintaudio_dma8.h"
#include "SDL_mintaudio_mcsn.h"

#ifndef MODE_MONO8
#define MODE_MONO8		MODE_MONO
#endif

/* Extended modes. Only available with SND_EXT */

#ifndef MODE_MONO16
#define MODE_MONO16		3	/* 16 bit mono, big-endian */
#define MODE_STEREO24	4	/* 24 bit stereo, big-endian */
#define MODE_STEREO32	5	/* 32 bit stereo, big-endian */
#define MODE_MONO24		6	/* 24 bit mono, big-endian */
#define MODE_MONO32		7	/* 32 bit mono, big-endian */
#endif
#ifndef SETSMPFREQ
#define	SETSMPFREQ	7
#endif

/*--- Defines ---*/

#define MINT_AUDIO_DRIVER_NAME "mint_xbios"

/* Debug print info */
#define DEBUG_NAME "audio:xbios: "
#if 0
#define DEBUG_PRINT(what) \
	{ \
		printf what; \
	}
#else
#define DEBUG_PRINT(what)
#endif

/*--- Static variables ---*/

static long cookie_snd;

/*--- Audio driver functions ---*/

static int Audio_Available(void)
{
	const char *envr = getenv("SDL_AUDIODRIVER");

	/* Check if user asked a different audio driver */
	if (envr && strcmp(envr, MINT_AUDIO_DRIVER_NAME) != 0)
	{
		DEBUG_PRINT((DEBUG_NAME "user asked a different audio driver\n"));
		return 0;
	}

	/* Cookie _SND present ? if not, assume ST machine */
	if (Getcookie(C__SND, &cookie_snd) != C_FOUND)
	{
		cookie_snd = SND_PSG;
	}

	/* Check if we have 16 bits audio */
	if ((cookie_snd & SND_16BIT) == 0)
	{
		DEBUG_PRINT((DEBUG_NAME "no 16 bits sound\n"));
		return 0;
	}

	/* Check if audio is lockable */
	if (Locksnd() != 1)
	{
		DEBUG_PRINT((DEBUG_NAME "audio locked by other application\n"));
		return 0;
	}

	Unlocksnd();

	DEBUG_PRINT((DEBUG_NAME "XBIOS audio available!\n"));
	return 1;
}

static void Audio_DeleteDevice(SDL_AudioDevice *device)
{
	SDL_free(device->hidden);
	SDL_free(device);
}

static long Mint_StopReplay(void)
{
	/* Stop replay */
	DMAAUDIO_IO.control = 0;
	return 0;
}

static long Mint_StartReplay(void)
{
	/* Start replay */
	DMAAUDIO_IO.control = SB_PLA_ENA | SB_PLA_RPT;
	return 0;
}

static void stop_replay(void)
{
	if ((short)Buffoper(0) == 0x88) /* XBIOS function not available */
		Supexec(Mint_StopReplay);
}

static void start_replay(void)
{
	if ((short)Buffoper(SB_PLA_ENA | SB_PLA_RPT) == 0x88) /* XBIOS function not available */
		Supexec(Mint_StartReplay);
}

static void Mint_LockAudio(_THIS)
{
	/* Stop replay */
	stop_replay();
}

static void Mint_UnlockAudio(_THIS)
{
	/* Restart replay */
	start_replay();
}

static void Mint_CloseAudio(_THIS)
{
	/* Stop replay */
	stop_replay();

	/* Uninstall interrupt */
	Jdisint(MFP_DMASOUND);

	SDL_MintAudio_FreeBuffers();

	/* Unlock sound system */
	Unlocksnd();
}

static void Mint_SwapBuffers(Uint8 *nextbuf, int nextsize)
{
	Setbuffer(SR_PLAY, nextbuf, nextbuf + nextsize);
}

/* Falcon XBIOS implementation of Devconnect() is buggy with external clock */
static void Devconnect2(int src, int dst, int sclk, int pre)
{		
	static const unsigned short MASK1[3] = { 0, 0x6000, 0 };
	static const unsigned short MASK2[4] = { 0xFFF0, 0xFF8F, 0xF0FF, 0x0FFF };
	static const unsigned short INDEX1[4] = {  1, 3, 5, 7 };
	static const unsigned short INDEX2[4] = {  0, 2, 4, 6 };
	unsigned short sync_div,dev_ctrl,dest_ctrl;
	void *oldstack;

	if (dst==0) {
		return;
	}

	oldstack=(void *)Super(0);

	dev_ctrl = DMAAUDIO_IO.dev_ctrl;
	dest_ctrl = DMAAUDIO_IO.dest_ctrl;
	dev_ctrl &= MASK2[src];

	if (src==ADC) {
		dev_ctrl |= MASK1[sclk];
	} else {
		dev_ctrl |= (INDEX1[sclk] << (src<<4));
	}

	if (dst & DMAREC) {		
		dest_ctrl &= 0xFFF0;
		dest_ctrl |= INDEX1[src];
	}

	if (dst & DSPRECV) {		
		dest_ctrl &= 0xFF8F;
		dest_ctrl |= (INDEX1[src]<<4); 
	}

	if (dst & EXTOUT) {		
		dest_ctrl &= 0xF0FF;
		dest_ctrl |= (INDEX1[src]<<8); 
	}

	if (dst & DAC) {		
		dev_ctrl &= 0x0FFF;
		dev_ctrl |= MASK1[sclk]; 
		dest_ctrl &=  0x0FFF;
		dest_ctrl |= (INDEX2[src]<<12); 
	}

	sync_div = DMAAUDIO_IO.sync_div;
	if (sclk==CLKEXT) {
		pre<<=8;
		sync_div &= 0xF0FF;
	} else {
		sync_div &= 0xFFF0;
	}
	sync_div |= pre;

	DMAAUDIO_IO.dev_ctrl = dev_ctrl;
	DMAAUDIO_IO.dest_ctrl = dest_ctrl;
	DMAAUDIO_IO.sync_div = sync_div;

	SuperToUser(oldstack);
}

static void Mint_CheckExternalClock(_THIS)
{
#define SIZE_BUF_CLOCK_MEASURE (44100/10)

	char *buffer;
	int i, j;

	/* DSP present with its GPIO port ? */
	if ((cookie_snd & SND_DSP)==0) {
		return;
	}

	buffer = Atari_SysMalloc(SIZE_BUF_CLOCK_MEASURE, MX_STRAM);
	if (buffer==NULL) {
		DEBUG_PRINT((DEBUG_NAME "Not enough memory for the measure\n"));
		return;
	}
	SDL_memset(buffer, 0, SIZE_BUF_CLOCK_MEASURE);

	Buffoper(0);
	Settracks(0, 0);
	Setmontracks(0);
	Setmode(MONO8);
	Jdisint(MFP_TIMERA);

	for (i=0; i<2; i++) {
		Gpio(GPIO_SET,7);      /* DSP port gpio outputs */
		Gpio(GPIO_WRITE,2+i);  /* 22.5792/24.576 MHz for 44.1/48KHz */
		Devconnect2(DMAPLAY, DAC, CLKEXT, CLK50K);  /* Matrix and clock source */
		Setbuffer(SR_PLAY, buffer, buffer + SIZE_BUF_CLOCK_MEASURE);		           /* Set buffer */
		Xbtimer(XB_TIMERA, 5, 38, SDL_MintAudio_XbiosInterruptMeasureClock); /* delay mode timer A, prediv /64, 1KHz */
		Jenabint(MFP_TIMERA);
		SDL_MintAudio_clocktics = 0;
		Buffoper(SB_PLA_ENA);
		usleep(110000);

		if((Buffoper(-1) & 1)==0) {
			if (SDL_MintAudio_clocktics) {
				unsigned long khz;

				khz = ((SIZE_BUF_CLOCK_MEASURE/SDL_MintAudio_clocktics) +1) & 0xFFFFFFFE;
				DEBUG_PRINT((DEBUG_NAME "measure %d: freq=%lu KHz\n", i+1, khz));

				if(khz==44) {
					for (j=1; j<4; j++) {
						SDL_MintAudio_AddFrequency(this, MASTERCLOCK_44K/(MASTERPREDIV_FALCON*(1<<j)), MASTERCLOCK_44K, (1<<j)-1, 2+i);
					}
				} else if (khz==48) {
					for (j=1; j<4; j++) {
						SDL_MintAudio_AddFrequency(this, MASTERCLOCK_48K/(MASTERPREDIV_FALCON*(1<<j)), MASTERCLOCK_48K, (1<<j)-1, 2+i);
					}
				}
			} else {
				DEBUG_PRINT((DEBUG_NAME "No measure\n"));
			}
		} else {
			DEBUG_PRINT((DEBUG_NAME "No SDMA clock\n"));
		}

		Buffoper(0);             /* stop */
		Jdisint(MFP_TIMERA);     /* Uninstall interrupt */
	}

	Mfree(buffer);
}

static void Mint_InitAudio(_THIS, SDL_AudioSpec *spec)
{
	int channels_mode, prediv;

	/* Stop currently playing sound */
	stop_replay();

	/* Set replay tracks */
	Settracks(0, 0);
	Setmontracks(0);

	/* Select replay format */
	channels_mode = STEREO16;
	switch (spec->format & 0xff)
	{
	case 8:
		if (spec->channels == 2)
		{
			channels_mode = STEREO8;
		} else
		{
			channels_mode = MONO8;
		}
		break;
	}
	if (Setmode(channels_mode) < 0)
	{
		DEBUG_PRINT((DEBUG_NAME "Setmode() failed\n"));
	}

	prediv = MINTAUDIO_frequencies[MINTAUDIO_numfreq].predivisor;
	if (MINTAUDIO_frequencies[MINTAUDIO_numfreq].gpio_bits != -1)
	{
		Gpio(GPIO_SET, 7);		/* DSP port gpio outputs */
		Gpio(GPIO_WRITE, MINTAUDIO_frequencies[MINTAUDIO_numfreq].gpio_bits);
		Devconnect2(DMAPLAY, DAC|EXTOUT, CLKEXT, prediv);
	} else
	{
		Devconnect2(DMAPLAY, DAC, CLK25M, prediv);
	}

	/* Set buffer */
	Mint_SwapBuffers(MINTAUDIO_audiobuf[0], MINTAUDIO_audiosize);
	
	/* Install interrupt */
	Jdisint(MFP_DMASOUND);
	Xbtimer(XB_TIMERA, 8, 1, SDL_MintAudio_XbiosInterrupt);
	Jenabint(MFP_DMASOUND);

	if (Setinterrupt(SI_TIMERA, SI_PLAY) < 0)
	{
		DEBUG_PRINT((DEBUG_NAME "Setinterrupt() failed\n"));
	}

	/* Go */
	start_replay();
	DEBUG_PRINT((DEBUG_NAME "hardware initialized\n"));
}

static int Mint_CheckAudio(_THIS, SDL_AudioSpec *spec)
{
	int i;

	DEBUG_PRINT((DEBUG_NAME "asked: %d bits, ", spec->format & 0x00ff));
	DEBUG_PRINT(("signed=%d, ", ((spec->format & 0x8000) != 0)));
	DEBUG_PRINT(("big endian=%d, ", ((spec->format & 0x1000) != 0)));
	DEBUG_PRINT(("channels=%d, ", spec->channels));
	DEBUG_PRINT(("freq=%d\n", spec->freq));

    if (spec->channels > 2)
    {
		spec->channels = 2;				/* no more than stereo! */
    }

	spec->format |= 0x8000;	/* Audio is always signed */
	if ((spec->format & 0x00ff) == 16)
	{
		spec->format |= 0x1000;	/* Audio is always big endian */
		spec->channels=2;	/* 16 bits always stereo */
	}

	MINTAUDIO_freqcount = 0;

	/* Add external clocks if present */
	Mint_CheckExternalClock(this);

	/* Standard clocks */
	SDL_MintAudio_AddFrequency(this, 49170, MASTERCLOCK_FALCON1, CLK50K, -1);
	SDL_MintAudio_AddFrequency(this, 32780, MASTERCLOCK_FALCON1, CLK33K, -1);
	SDL_MintAudio_AddFrequency(this, 24585, MASTERCLOCK_FALCON1, CLK25K, -1);
	SDL_MintAudio_AddFrequency(this, 19668, MASTERCLOCK_FALCON1, CLK20K, -1);
	SDL_MintAudio_AddFrequency(this, 16390, MASTERCLOCK_FALCON1, CLK16K, -1);
	/* 14049 hz invalid for CODEC */
	SDL_MintAudio_AddFrequency(this, 12292, MASTERCLOCK_FALCON1, CLK12K, -1);
	/* 10927 hz invalid for CODEC */
	SDL_MintAudio_AddFrequency(this, 9834, MASTERCLOCK_FALCON1, CLK10K, -1);
	/* 8940 hz invalid for CODEC */
	SDL_MintAudio_AddFrequency(this, 8195, MASTERCLOCK_FALCON1, CLK8K, -1);

#if 1
	for (i=0; i<MINTAUDIO_freqcount; i++)
	{
		DEBUG_PRINT((DEBUG_NAME "freq %d: %lu Hz, clock %lu, prediv %d\n",
			i, MINTAUDIO_frequencies[i].frequency, MINTAUDIO_frequencies[i].masterclock,
			MINTAUDIO_frequencies[i].predivisor
		));
	}
#endif

	MINTAUDIO_numfreq = SDL_MintAudio_SearchFrequency(this, spec->freq);
	spec->freq = MINTAUDIO_frequencies[MINTAUDIO_numfreq].frequency;

	DEBUG_PRINT((DEBUG_NAME "obtained: %d bits, ", spec->format & 0x00ff));
	DEBUG_PRINT(("signed=%d, ", ((spec->format & 0x8000) != 0)));
	DEBUG_PRINT(("big endian=%d, ", ((spec->format & 0x1000) != 0)));
	DEBUG_PRINT(("channels=%d, ", spec->channels));
	DEBUG_PRINT(("freq=%d\n", spec->freq));

	return 0;
}

static int Mint_OpenAudio(_THIS, SDL_AudioSpec *spec)
{
	short locked;

	/* Lock sound system */
	locked = (short)Locksnd();
	if (locked != 1 && locked != 0x80) /* XBIOS function not available */
	{
   	    SDL_SetError("Mint_OpenAudio: Audio system already in use");
		return -1;
	}

	SDL_MintAudio_device = this;

	/* Check audio capabilities */
	if (Mint_CheckAudio(this, spec) < 0)
	{
		Unlocksnd();
		return -1;
	}

	if (!SDL_MintAudio_InitBuffers(spec))
	{
		Unlocksnd();
		return -1;
	}

	/* Setup audio hardware */
	MINTAUDIO_swapbuf = Mint_SwapBuffers;
	Mint_InitAudio(this, spec);

	return 1;							/* We don't use SDL threaded audio */
}

static SDL_AudioDevice *Audio_CreateDevice(int devindex)
{
	SDL_AudioDevice *this;

	/* Initialize all variables that we clean on shutdown */
	this = (SDL_AudioDevice *)SDL_calloc(1, sizeof(*this));
    if (this)
    {
        this->hidden = (struct SDL_PrivateAudioData *) SDL_calloc(1, sizeof(*this->hidden));
    }
    if (this == NULL || this->hidden == NULL)
    {
        SDL_OutOfMemory();
		SDL_free(this);
        return 0 ;
    }

    /* Set the function pointers */
    this->OpenAudio = Mint_OpenAudio;
    this->CloseAudio = Mint_CloseAudio;
    this->LockAudio = Mint_LockAudio;
    this->UnlockAudio = Mint_UnlockAudio;
    this->free = Audio_DeleteDevice;

    return this;
}

AudioBootStrap MINTAUDIO_XBIOS_bootstrap = {
	MINT_AUDIO_DRIVER_NAME, "MiNT XBIOS audio driver", Audio_Available, Audio_CreateDevice
};
